import sys

import demeter_fetch.common.utils as utils
from demeter_fetch.core import get_commend_args, download
from demeter_fetch.tools import date_to_height

if __name__ == "__main__":
    if len(sys.argv) == 1:
        utils.print_log("use parameter -h for help")
        exit(1)
    if len(sys.argv) > 1 and sys.argv[1].endswith(".toml"):
        download(sys.argv[1])
        exit(0)

    args = get_commend_args()
    if args.config is not None:
        download(args.config)
    elif args.tools is not None:
        if args.tools == "date_to_height":
            date_to_height(args)
        pass



from datetime import date
from typing import Dict

import pandas as pd
import demeter_fetch.processor_aave.aave_utils as aave_utils
from demeter_fetch.common import AaveDailyNode, get_tx_type, AaveNodesNames, KECCAK
from demeter_fetch.common.nodes import AaveDailyParam


class AaveTick(AaveDailyNode):
    def __init__(self, depends):
        super().__init__(depends)
        self.name = AaveNodesNames.tick

    def _get_file_name(self, param: AaveDailyParam) -> str:
        return f"{self.from_config.chain.name}-aave_v3-{param.token}-{param.day.strftime('%Y-%m-%d')}.tick.csv"

    def _process_one_day(self, data: Dict[str, Dict[str, pd.DataFrame]], day: date, tokens) -> Dict[str, pd.DataFrame]:
        ret: Dict[str, pd.DataFrame] = {}
        df = data[AaveNodesNames.raw]
        for token, token_df in df.items():
            token_df["tx_type"] = token_df.apply(lambda x: get_tx_type(x.topics), axis=1)
            token_df = token_df[
                token_df["tx_type"].isin(
                    [
                        KECCAK.AAVE_REPAY,
                        KECCAK.AAVE_BORROW,
                        KECCAK.AAVE_SUPPLY,
                        KECCAK.AAVE_WITHDRAW,
                        KECCAK.AAVE_LIQUIDATION,
                    ]
                )
            ]
            ret[token] = preprocess_one(token_df)
        return ret


def preprocess_one(df: pd.DataFrame):
    append_columns = ["reserve", "owner", "amount", "liquidator", "debt_asset", "debt_amount", "atoken"]
    ret_df = pd.DataFrame()
    if df.empty:
        for column in df.columns:
            ret_df[column] = None
        return ret_df
    ret_df[append_columns] = df.apply(
        lambda x: aave_utils.handle_event(x.tx_type, x.topics, x.data),
        axis=1,
        result_type="expand",
    )
    ret_df[["block_number", "transaction_hash", "block_timestamp", "transaction_index", "log_index", "tx_type"]] = df[
        ["block_number", "transaction_hash", "block_timestamp", "transaction_index", "log_index", "tx_type"]
    ]
    ret_df["tx_type"] = ret_df.apply(lambda x: x.tx_type.name, axis=1)
    final_order = [
        "block_number",
        "block_timestamp",
        "tx_type",
        "transaction_hash",
        "transaction_index",
        "log_index",
        "reserve",
        "owner",
        "amount",
        "liquidator",
        "debt_asset",
        "debt_amount",
        "atoken",
    ]
    ret_df = ret_df[final_order]

    return ret_df



from _decimal import Decimal
from typing import List

import numpy as np

import demeter_fetch.common._typing as TYPE

RAY = 10**27


def decode_event_ReserveDataUpdated(row):
    return (
        Decimal(int(row["DATA"][2 : (2 + 64)], 16)) / RAY,
        Decimal(int(row["DATA"][(2 + 64) : (2 + 64 * 2)], 16)) / RAY,
        Decimal(int(row["DATA"][(2 + 64 * 2) : (2 + 64 * 3)], 16)) / RAY,
        Decimal(int(row["DATA"][(2 + 64 * 3) : (2 + 64 * 4)], 16)) / RAY,
        Decimal(int(row["DATA"][(2 + 64 * 4) :], 16)) / RAY,
    )


def split_topic(value: str) -> List[str]:
    value = value.strip("[]").replace('"', "").replace("'", "").replace(" ", "").replace("\n", ",")
    return value.split(",")


def hex_to_address(topic_str):
    return "0x" + topic_str[26:]


def signed_int(h):
    """
    Converts hex values to signed integers.
    """
    s = bytes.fromhex(h[2:])
    i = int.from_bytes(s, "big", signed=True)
    return i


def handle_event(tx_type, topics_str, data_hex):
    reserve = None
    owner = None
    amount = None
    debt_asset = None
    debt_amount = Decimal(np.nan)
    liquidator = None
    atoken = None

    topic_list = split_topic(topics_str)
    no_0x_data = data_hex[2:]
    chunk_size = 64
    chunks = len(no_0x_data)
    match tx_type:
        # reserve,
        case TYPE.KECCAK.AAVE_SUPPLY:
            # reserve,onBehalfOf,referralCode,user,amount
            # reserve: trading pool
            # onBehalfOf: supply asset owner
            # referralCode: unique code for 3rd party referral program integration. Use 0 for no referral
            # user: msg sender
            # amount: supplied asset amount
            reserve = hex_to_address(topic_list[1])
            owner = onBehalfOf = hex_to_address(topic_list[2])
            referralCode = signed_int(topic_list[3])
            split_data = ["0x" + no_0x_data[i : i + chunk_size] for i in range(0, chunks, chunk_size)]
            user = hex_to_address(split_data[0])
            amount = signed_int(split_data[1])
        case TYPE.KECCAK.AAVE_WITHDRAW:
            # reserve,user,to,amount
            # reserve: trading pool
            # user: msg sender
            # to: receiver
            # amount: asset amount
            reserve = hex_to_address(topic_list[1])
            user = hex_to_address(topic_list[2])
            owner = to = hex_to_address(topic_list[3])
            amount = signed_int(data_hex)
        case TYPE.KECCAK.AAVE_BORROW:
            # reserve,onBehalfOf,referralCode,user,amount,interestRateMode,borrowRate
            # reserve: trading pool
            # onBehalfOf: borrow asset owner
            # referralCode: refer code
            # user: msg sender
            # amount: borrow asset amount
            # interestRateMode: The rate mode: 1 for Stable, 2 for Variable
            # borrowRate: The numeric rate at which the user has borrowed, expressed in ray
            reserve = hex_to_address(topic_list[1])
            owner = onBehalfOf = hex_to_address(topic_list[2])
            referralCode = signed_int(topic_list[3])
            split_data = ["0x" + no_0x_data[i : i + chunk_size] for i in range(0, chunks, chunk_size)]
            user = hex_to_address(split_data[0])
            amount = signed_int(split_data[1])
            interest_rate_mode = signed_int(split_data[2])
            borrow_rate = signed_int(split_data[3])
        case TYPE.KECCAK.AAVE_REPAY:
            # repayer 还款人, user 债务人
            # reserve,user,repayer,amount,useATokens
            # reserve: trading pool
            # repayer: some body pay directly for repayment
            # user: The beneficiary of the repayment, getting his debt reduced
            # amount: repay amount
            # useATokens: useATokens True if the repayment is done using aTokens, `false` if done with underlying asset directly
            reserve = hex_to_address(topic_list[1])
            owner = user = hex_to_address(topic_list[2])
            repayer = hex_to_address(topic_list[3])
            split_data = ["0x" + no_0x_data[i : i + chunk_size] for i in range(0, chunks, chunk_size)]
            amount = signed_int(split_data[0])
            atoken = signed_int(split_data[1])
        case TYPE.KECCAK.AAVE_LIQUIDATION:
            # liquidator repay debt_asset debt_amount and get collateral_asset liquidated_collateral_amount
            # 清算者购买被清算者一定数量的抵押资产（collateral asset）来偿还被清算者特定资产债务（debt asset）
            # 清算者将用于购买的债务资产转移到atoken，atoken将抵押物或者抵押物对应的atoken转移到清算者，销毁被清算者的一部分债务代币。
            # collateralAsset,debtAsset,user,debtToCover,liquidatedCollateralAmount,liquidator,receiveAToken
            # collateralAsset: The address of the underlying asset used as collateral, to receive as result of the liquidation
            # debtAsset: The address of the underlying borrowed asset to be repaid with the liquidation
            # user: The address of the borrower getting liquidated
            # debtToCover: The debt amount of borrowed `asset` the liquidator wants to cover
            # liquidatedCollateralAmount: The amount of collateral received by the liquidator
            # liquidator: The address of the liquidator
            # receiveAToken: True if the liquidators wants to receive the collateral aTokens, `false` if he wants to receive the underlying collateral asset directly
            split_data = ["0x" + no_0x_data[i : i + chunk_size] for i in range(0, chunks, chunk_size)]
            reserve = collateral_asset = hex_to_address(topic_list[1])  # 清算者想要购买的抵押资产。
            owner = user = hex_to_address(topic_list[3])  # 被清算者
            amount = liquidated_collateral_amount = signed_int(split_data[1])  # 清算者购买的抵押资产数量
            liquidator = hex_to_address(split_data[2])  # 清算者
            debt_asset = hex_to_address(topic_list[2])  # 被清算的债务资产，同时也是清算者要支付的资产。
            debt_amount = debtToCover = signed_int(split_data[0])  # 替被清算者偿还的贷款额
            atoken = signed_int(split_data[3])  # 清算者是否以atoken的形式接收清算得到的抵押资产。
        case _:
            raise ValueError("not support tx type")
    return (reserve, owner, Decimal(amount), liquidator, debt_asset, Decimal(debt_amount), atoken)



import time
from dataclasses import dataclass
from datetime import date, datetime
from typing import List, Tuple, Dict

import demeter_fetch.common.utils as utils
from .. import ChainType


@dataclass
class ContractConfig:
    address: str
    topics: List[str]


height_cache: Dict[date, Tuple[int, int]] = {}


def get_height_from_date(
    day: date, chain: ChainType, http_proxy, etherscan_api_key, sleep_seconds=1, sleep_seconds_without_key=8
) -> (int, int):
    if day in height_cache.keys():
        return height_cache[day]
    utils.print_log(f"Query height range in {day}")
    if etherscan_api_key is None:
        sleep_seconds = sleep_seconds_without_key
    start_height = utils.ApiUtil.query_blockno_from_time(
        chain, datetime.combine(day, datetime.min.time()), False, http_proxy, etherscan_api_key
    )
    utils.print_log(f"Querying end timestamp, wait for {sleep_seconds} seconds to prevent max rate limit")
    time.sleep(sleep_seconds)  # to prevent request limit
    end_height = utils.ApiUtil.query_blockno_from_time(
        chain, datetime.combine(day, datetime.max.time()), True, http_proxy, etherscan_api_key
    )
    height_cache[day] = (start_height, end_height)
    return start_height, end_height



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 2024-01-10 11:08
# @Author  : 32ethers
# @Description:
from datetime import date
from typing import List

import pandas as pd

from .big_query_utils import BigQueryChain, query_by_sql
from .. import ChainTypeConfig
from ..common import FromConfig, utils, KECCAK


def _update_df(df: pd.DataFrame) -> pd.DataFrame:
    df["topics"] = df["topics"].apply(lambda x: x.tolist())
    df = df.sort_values(["block_number", "log_index"], ascending=[True, True])
    df["block_timestamp"] = df["block_timestamp"].dt.tz_localize(None)
    df["block_timestamp"] = pd.Series(df["block_timestamp"].dt.to_pydatetime(), index=df.index, dtype=object)
    return df


def bigquery_pool(config: FromConfig, day: date):
    day_str = day.strftime("%Y-%m-%d")
    sql = f"""
    SELECT block_number,block_timestamp, transaction_hash , transaction_index , log_index, topics , DATA as data
        FROM {BigQueryChain[config.chain.value].value["table_name"]}
        WHERE  topics[SAFE_OFFSET(0)] in {KECCAK.SWAP.value, KECCAK.BURN.value, KECCAK.COLLECT.value, KECCAK.MINT.value}
            AND DATE(block_timestamp) =  DATE("{day_str}") AND address = "{config.uniswap_config.pool_address}"
    """
    df = query_by_sql(sql, config.big_query.auth_file, config.http_proxy)
    df = _update_df(df)
    return df


def bigquery_proxy_lp(config: FromConfig, day: date):
    day_str = day.strftime("%Y-%m-%d")
    sql = f"""
    SELECT block_number,block_timestamp, transaction_hash , transaction_index , log_index, topics , DATA as data
        FROM {BigQueryChain[config.chain.value].value["table_name"]}
        WHERE  topics[SAFE_OFFSET(0)] in {KECCAK.UNI_PROXY_INCREASE.value, KECCAK.UNI_PROXY_DECREASE.value, KECCAK.UNI_PROXY_COLLECT.value}
            AND DATE(block_timestamp) =  DATE("{day_str}") AND address = "{ChainTypeConfig[config.chain]["uni_proxy_addr"]}"
    """
    df = query_by_sql(sql, config.big_query.auth_file, config.http_proxy)
    df = _update_df(df)
    return df


def bigquery_proxy_transfer(config: FromConfig, day: date):
    day_str = day.strftime("%Y-%m-%d")
    sql = f"""
    SELECT block_number,block_timestamp, transaction_hash , transaction_index , log_index, topics , DATA as data
        FROM {BigQueryChain[config.chain.value].value["table_name"]}
        WHERE  topics[SAFE_OFFSET(0)] in ('{KECCAK.TRANSFER.value}')
            AND DATE(block_timestamp) =  DATE("{day_str}") AND address = "{ChainTypeConfig[config.chain]["uni_proxy_addr"]}"
    """
    df = query_by_sql(sql, config.big_query.auth_file, config.http_proxy)
    df["topics"] = df["topics"].apply(lambda x: x.tolist())
    return df


def bigquery_transaction(config: FromConfig, day: date, tx: List[str]):
    day_str = day.strftime("%Y-%m-%d")
    tx_str = ",".join(['"' + utils.hex_to_length(x, 64) + '"' for x in tx])
    sql = f"""
    select `hash` as transaction_hash ,block_number,transaction_index,from_address as `from`,to_address as `to`,value
    from {BigQueryChain[config.chain.value].value["tx_table_name"]}
    where DATE(block_timestamp) = DATE("{day_str}")
        and  `hash` in ({tx_str})
    """
    df = query_by_sql(sql, config.big_query.auth_file, config.http_proxy)
    return df


def bigquery_aave(config: FromConfig, day: date, tokens: List[str]):
    day_str = day.strftime("%Y-%m-%d")
    token_str = ",".join(['"' + utils.hex_to_length(x, 64) + '"' for x in tokens])
    keccak_str = ",".join(
        [
            '"' + utils.hex_to_length(x, 64) + '"'
            for x in [
                KECCAK.AAVE_SUPPLY.value,
                KECCAK.AAVE_REPAY.value,
                KECCAK.AAVE_BORROW.value,
                KECCAK.AAVE_WITHDRAW.value,
                KECCAK.AAVE_LIQUIDATION.value,
                KECCAK.AAVE_UPDATED.value,
            ]
        ]
    )
    sql = f"""
            SELECT block_number,block_timestamp,transaction_hash,transaction_index,log_index,topics,DATA as data
            FROM {BigQueryChain[config.chain.value].value["table_name"]}
            WHERE
              topics[SAFE_OFFSET(0)] IN ({keccak_str})
              AND topics[SAFE_OFFSET(1)] IN ({token_str})
              AND DATE(block_timestamp) >= DATE("{day_str}")
              AND DATE(block_timestamp) <= DATE("{day_str}")
              AND address = "{ChainTypeConfig[config.chain]['aave_v3_pool_addr']}"
        """
    df = query_by_sql(sql, config.big_query.auth_file, config.http_proxy)
    df = _update_df(df)
    return df



import json
import os.path
import pickle
import random
import time
from concurrent.futures import ThreadPoolExecutor, as_completed
from dataclasses import dataclass
from datetime import datetime
from operator import itemgetter
from typing import List, Dict

import numpy as np
import pandas as pd
import requests
from tqdm import tqdm  # process bar

import demeter_fetch.common.utils as utils
from demeter_fetch.common._typing import ChainType
from demeter_fetch.common._typing import EthError
from .source_utils import ContractConfig


@dataclass
class GetLogsParam:
    address: str
    fromBlock: int
    toBlock: int
    topics: List[str] | None


class EthRpcClient:
    def __init__(self, endpoint: str, proxy="", auth=""):
        self.session = requests.Session()
        adapter = requests.adapters.HTTPAdapter(pool_connections=5, pool_maxsize=20)
        self.session.mount("https://", adapter)
        self.session.mount("http://", adapter)
        self.headers = {}
        self.endpoint = endpoint
        if auth:
            self.headers["Authorization"] = auth
        self.proxies = (
            {
                "http": proxy,
                "https": proxy,
            }
            if proxy
            else {}
        )

    def __del__(self):
        self.session.close()

    @staticmethod
    def __encode_json_rpc(method: str, params: list):
        return {"jsonrpc": "2.0", "method": method, "params": params, "id": random.randint(1, 2147483648)}

    @staticmethod
    def __decode_json_rpc(response: requests.Response):
        try:
            content = response.json()
        except Exception as e:
            print(f"Decode rpc response failed, error: {e}")
            raise e

        if "error" in content:
            raise EthError(content["error"]["code"], content["error"]["message"])
        return content["result"]

    def do_post(self, param):
        return self.session.post(self.endpoint, json=param, proxies=self.proxies, headers=self.headers)

    def get_block(self, height):
        return self.send("eth_getBlockByNumber", [hex(height), False])

    def get_block_timestamp(self, height):
        resp = self.get_block(height)
        if resp:
            timestamp = int(resp["timestamp"], 16)
            return datetime.utcfromtimestamp(timestamp)
        else:
            return None

    def get_tx_receipt(self, tx_hash):
        return self.send("eth_getTransactionReceipt", [tx_hash])

    def get_tx(self, tx_hash):
        """
        response:
            "hash":"0xc6ef2fc5426d6ad6fd9e2a26abeab0aa2411b7ab17f30a99d3cb96aed1d1055b",
            "nonce":"0x",
            "blockHash": "0xbeab0aa2411b7ab17f30a99d3cb9c6ef2fc5426d6ad6fd9e2a26a6aed1d1055b",
            "blockNumber": "0x15df", // 5599
            "transactionIndex":  "0x1", // 1
            "from":"0x407d73d8a49eeb85d32cf465507dd71d507100c1",
            "to":"0x85h43d8a49eeb85d32cf465507dd71d507100c1",
            "value":"0x7f110", // 520464
            "gas": "0x7f110", // 520464
            "gasPrice":"0x09184e72a000",
            "input":"0x603880600c6000396000f300603880600c6000396000f3603880600c6000396000f360",
        """
        return self.send("eth_getTransactionByHash", [tx_hash])

    def get_logs(self, param: GetLogsParam):
        if param.toBlock:
            param.toBlock = hex(param.toBlock)
        if param.fromBlock:
            param.fromBlock = hex(param.fromBlock)
        return self.send("eth_getLogs", [vars(param)])

    def send(self, commend: str, params: List):
        response = self.do_post(EthRpcClient.__encode_json_rpc(commend, params))
        return EthRpcClient.__decode_json_rpc(response)


class HeightCacheManager:
    """
    高度缓存
    """

    height_cache_file_name = "_height_timestamp.pkl"

    def __init__(self, chain: ChainType, save_path: str):
        self.height_cache_path = os.path.join(save_path, chain.value + HeightCacheManager.height_cache_file_name)
        if os.path.exists(self.height_cache_path):
            with open(self.height_cache_path, "rb") as f:
                self.block_dict = pickle.load(f)
                utils.print_log(f"Height cache has loaded, length: {len(self.block_dict)}")
        else:
            self.block_dict: Dict[int, datetime] = {}
            utils.print_log("Can not find a height cache, will generate one")

    def has(self, height: int):
        return height in self.block_dict

    def get(self, height: int):
        if height in self.block_dict:
            return self.block_dict[height]
        else:
            return None

    def set(self, height: int, timestamp: datetime):
        self.block_dict[height] = timestamp

    def save(self):
        with open(self.height_cache_path, "wb") as f:
            pickle.dump(self.block_dict, f)
        utils.print_log(f"Save block timestamp cache to {self.height_cache_path}, length: {len(self.block_dict)}")


def _query_tx_receipt(param):
    tx_hash, client = param
    resp = client.get_tx_receipt(tx_hash)
    return resp


def _query_tx(param):
    tx_hash, client = param
    resp = client.get_tx(tx_hash)
    return resp


def query_tx(client: EthRpcClient, tx_list: pd.Series, threads=10) -> pd.DataFrame:
    param_list = [(tx_hash, client) for tx_hash in tx_list]
    with ThreadPoolExecutor(max_workers=threads) as executor:
        txes = list(tqdm(executor.map(_query_tx, param_list), ncols=60, position=1, leave=False, total=len(tx_list)))

    tx_list = []
    for tx in txes:
        if tx is None:
            continue
        tx_list.append(
            {
                "transaction_hash": tx["hash"],  # "0xc6ef2fc5426d6ad6fd9e2a26abeab0aa2411b7ab17f30a99d3cb96aed1d1055b",
                "block_number": int(tx["blockNumber"], 16),
                "transaction_index": int(tx["transactionIndex"], 16),  #  "0x1",
                "from": tx["from"],  #  "0x407d73d8a49eeb85d32cf465507dd71d507100c1",
                "to": tx["to"],  #  "0x85h43d8a49eeb85d32cf465507dd71d507100c1",
                "value": int(tx["value"], 16),  #  "0x7f110",
                # "gas": int(tx["gas"], 16),  #  "0x7f110",
                # "gas_price": int(tx["gasPrice"], 16),  #  "0x09184e72a000",
            }
        )

    df = pd.DataFrame(tx_list)
    return df


def query_event_by_tx(client: EthRpcClient, tx_list: pd.Series, threads=10) -> pd.DataFrame:
    param_list = [(tx_hash, client) for tx_hash in tx_list]
    with ThreadPoolExecutor(max_workers=threads) as executor:
        tx_receipts = list(
            tqdm(executor.map(_query_tx_receipt, param_list), ncols=60, position=1, leave=False, total=len(tx_list))
        )

    logs_list = []
    for tx in tx_receipts:
        if tx is None:
            continue
        for log in tx["logs"]:
            logs_list.append(
                {
                    "block_number": int(log["blockNumber"], 16),
                    "transaction_hash": log["transactionHash"],
                    "transaction_index": int(log["transactionIndex"], 16),
                    "from": tx["from"],
                    "to": tx["to"],
                    "log_index": int(log["logIndex"], 16),
                    "log_address": log["address"],
                    "topics": log["topics"],
                    "data": log["data"] if log["data"] != "0x" else "",
                }
            )
    df = pd.DataFrame(logs_list)
    return df


def query_event_by_height(
    chain: ChainType,
    client: EthRpcClient,
    contract_config: ContractConfig,
    start_height: int,
    end_height: int,
    height_cache: HeightCacheManager = None,
    save_path: str = "./",
    save_every_query: int = 10,
    batch_size: int = 500,
    one_by_one: bool = False,
    skip_timestamp: bool = False,
) -> List[str]:
    """
    根据输入参数, 下载对应高度的log,
    log会按照高度划分, 保存为临时文件.

    :param chain: chain
    :param client: eth rpc client
    :param contract_config: 配置信息
    :param start_height: 开始高度
    :param end_height: 结束高度
    :param height_cache: 高度-时间戳 缓存的路径
    :param save_path: 临时文件保存路径
    :param save_every_query: 多少次查询会保存一次临时文件. save_every_query * batch_size = 临时文件里的数据条数.
    :param batch_size: 一次下载多少个block的log
    :param one_by_one: 逐个下载每一个topic, 还是下载所有的topic再筛选掉不需要的.
    :param skip_timestamp: skip query block timestamp
    :return: 临时文件的文件名
    :rtype:
    """

    collect_dt, logs_to_save, collect_start = None, [], None  # collect date, date log by day，collect start time
    tmp_file_full_path_list = []
    if not height_cache:
        height_cache = HeightCacheManager(chain, save_path)
    batch_count = start_blk = end_blk = 0
    skip_until = -1
    utils.print_log(f"Querying {contract_config.address} from {start_height} to {end_height}")
    with tqdm(total=(end_height - start_height + 1), ncols=60, position=1, leave=False) as pbar:
        for height_slice in _cut([i for i in range(start_height, end_height + 1)], batch_size):
            start = height_slice[0]
            end = height_slice[-1]
            if end <= skip_until:
                batch_count += 1
                pbar.update(n=len(height_slice))
                continue
            if batch_count % save_every_query == 0:
                start_blk = start
                tmp_end_blk = start + batch_size * save_every_query - 1
                if tmp_end_blk >= end_height:
                    tmp_end_blk = end_height
                # 下载之前检测文件是否已经存在, 如果存在跳过下载
                tmp_file_path = get_tmp_file_path(save_path, start_blk, tmp_end_blk, chain, contract_config.address)
                if os.path.exists(tmp_file_path):
                    skip_until = tmp_end_blk
                    batch_count += 1
                    pbar.update(n=len(height_slice))
                    tmp_file_full_path_list.append(tmp_file_path)
                    continue
            # 下载之前检测文件是否已经存在, 如果存在跳过下载

            if one_by_one:
                logs = []
                for topic_hex in contract_config.topics:
                    tmp_logs = client.get_logs(GetLogsParam(contract_config.address, start, end, [topic_hex]))
                    logs.extend(tmp_logs)
            else:
                logs = client.get_logs(GetLogsParam(contract_config.address, start, end, None))
            log_list = []
            for log in logs:
                log["blockNumber"] = int(log["blockNumber"], 16)
                if len(log["topics"]) > 0 and (log["topics"][0] in contract_config.topics):
                    if log["removed"]:
                        continue
                    # block_number, block_timestamp, transaction_hash, transaction_index, log_index, topics, data
                    log_list.append(
                        {
                            "block_number": log["blockNumber"],
                            "transaction_hash": log["transactionHash"],
                            "transaction_index": log["transactionIndex"],
                            "log_index": log["logIndex"],
                            "data": log["data"],
                            "topics": json.dumps(log["topics"]),
                        }
                    )
            for log in log_list:
                log["log_index"] = int(log["log_index"], 16)
                log["transaction_index"] = int(log["transaction_index"], 16)
            if skip_timestamp:
                logs_to_save.extend(log_list)
            else:
                with ThreadPoolExecutor(max_workers=10) as t:
                    obj_lst = []
                    for data in log_list:
                        obj = t.submit(_fill_block_info, data, client, height_cache)
                        obj_lst.append(obj)
                    for future in as_completed(obj_lst):
                        data = future.result()
                        logs_to_save.append(data)

            # if got enough blocks, save file
            batch_count += 1
            if batch_count % save_every_query == 0:
                # save tmp file
                logs_to_save = sorted(logs_to_save, key=itemgetter("block_number", "transaction_index", "log_index"))
                end_blk = end
                tmp_file_full_path_list.append(
                    save_tmp_file(save_path, logs_to_save, start_blk, end_blk, chain, contract_config.address)
                )
                logs_to_save = []
            pbar.update(n=len(height_slice))
    if batch_count % save_every_query != 0 and skip_until < end:  # save tail queries
        logs_to_save = sorted(logs_to_save, key=itemgetter("block_number", "transaction_index", "log_index"))
        end_blk = end
        tmp_file_full_path_list.append(
            save_tmp_file(save_path, logs_to_save, start_blk, end_blk, chain, contract_config.address)
        )
    height_cache.save()
    return tmp_file_full_path_list


def load_tmp_file(full_path) -> List:
    with open(full_path, "rb") as f:
        data = pickle.load(f)
    return data


def get_tmp_file_path(save_path, start, end, chain, address):
    return os.path.join(save_path, f"{chain.name}-{address}-{start}-{end}.tmp.pkl")


def save_tmp_file(save_path, logs, start, end, chain, address):
    file_path = get_tmp_file_path(save_path, start, end, chain, address)
    with open(file_path, "wb") as f:
        pickle.dump(logs, f)
    return file_path


def _cut(obj, sec):
    return [obj[i : i + sec] for i in range(0, len(obj), sec)]


def _fill_block_info(log, client: EthRpcClient, block_dict: HeightCacheManager):
    height = log["block_number"]
    if not block_dict.has(height):
        block_dt = client.get_block_timestamp(height)
        block_dict.set(height, block_dt)
    log["block_timestamp"] = block_dict.get(height).isoformat()
    log["block_dt"] = block_dict.get(height)

    return log


def set_position_id(row: pd.Series) -> str:
    if (row["position_id"] is not None) and (not np.isnan(row["position_id"])):
        return str(int(row["position_id"]))
    return f"{row['owner']}-{int(row['tick_lower'])}-{int(row['tick_upper'])}"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 2024-01-09 11:21
# @Author  : 32ethers
# @Description:
from collections import namedtuple
from datetime import timedelta, date

import os
from typing import List, Dict, Callable, Tuple

import pandas as pd
from tqdm import tqdm

from ._typing import Config, FromConfig
from .utils import TimeUtil, set_global_pbar

EmptyNamedTuple = namedtuple("EmptyNamedTuple", [])


class Node:
    def __init__(self, depends: List, name=""):
        self.name = name
        self.depends: List[Node] = depends
        self.depends_dict: Dict[str, Node] = {d.name: d for d in self.depends}
        self.config: Config | None = None
        self.from_config: FromConfig | None = None
        self.to_path: str | None = None

    def set_config(self, config: Config):
        self.config = config
        self.from_config = config.from_config
        self.to_path = config.to_config.save_path

    def work(self):
        set_global_pbar(None)
        missing_params: List[EmptyNamedTuple] = []
        if self.config.to_config.skip_existed:
            step_file_names = self.get_file_paths
            for key, fn in step_file_names.items():
                if not os.path.exists(fn):
                    missing_params.append(key)
                    break
            if len(missing_params) < 1:
                return
        data = {}
        for depend in self.depends:
            data[depend.name] = list(depend.get_file_paths.values())
        pbar = tqdm(total=len(missing_params), ncols=80, position=0, leave=False)
        set_global_pbar(pbar)
        for param in missing_params:
            df = self._process_one(data, param)
            df.to_csv(self.get_file_path(param), index=False)
            pbar.update()

    def _process_one(self, data: Dict[str, List[str]], param: namedtuple) -> pd.DataFrame:
        return pd.DataFrame()

    def _get_file_name(self, param: namedtuple) -> str:
        return f"{self.name}-{str(param)}.csv"

    def get_file_path(self, param: namedtuple) -> str:
        return os.path.join(self.to_path, self._get_file_name(param))

    @property
    def get_file_paths(self) -> Dict[namedtuple, str]:
        return {
            EmptyNamedTuple(): self.get_file_path(EmptyNamedTuple()),
        }

    @property
    def load_csv_converter(self) -> Dict[str, Callable]:
        """
        if you want to read csv generated by this step with pandas, you have to use those csv converter
        :return:
        """
        return {}

    def get_depend_by_name(self, name: str):
        return self.depends_dict[name]

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name


DailyParam = namedtuple("DailyParam", ["day"])


class DailyNode(Node):
    """
    Node whose input and output and dependings are daily, and generate only one file per day.
    """

    def __init__(self, depends: List):
        super().__init__(depends)

    def work(self):
        set_global_pbar(None)
        # if daily, global loop will handle processbar, outfile existence, gather param
        day_idx = self.from_config.start
        pbar = tqdm(total=(self.from_config.end - self.from_config.start).days + 1, ncols=80, position=0, leave=False)
        set_global_pbar(pbar)
        while day_idx <= self.from_config.end:
            # day_str = day_idx.strftime("%Y-%m-%d")
            day_param = DailyParam(day_idx)
            step_file_name = self.get_file_path(day_param)
            if self.config.to_config.skip_existed and os.path.exists(step_file_name):
                day_idx += timedelta(days=1)
                continue
            param = {}
            for depend in self.depends:
                depend_file_path = depend.get_file_path(day_param)
                param[depend.name] = pd.read_csv(depend_file_path, converters=depend.load_csv_converter)

            df = self._process_one_day(param, day_idx)
            df.to_csv(self.get_file_path(day_param), index=False)
            day_idx += timedelta(days=1)
            pbar.update()

    def _process_one_day(self, data: Dict[str, pd.DataFrame], day: date) -> pd.DataFrame:
        return pd.DataFrame()

    @property
    def get_file_paths(self) -> Dict[namedtuple, str]:
        return {
            DailyParam(day): self.get_file_path(DailyParam(day))
            for day in TimeUtil.get_date_array(self.from_config.start, self.from_config.end)
        }


AaveDailyParam = namedtuple("AaveDailyParam", ["day", "token"])


class AaveDailyNode(Node):
    """
    Node whose input and output and dependings are daily, and generate multiple files per day.
    """

    def __init__(self, depends: List):
        super().__init__(depends)

    @property
    def get_file_paths(self) -> Dict[namedtuple, str]:
        ret = {}
        for day in TimeUtil.get_date_array(self.from_config.start, self.from_config.end):
            for token in self.from_config.aave_config.tokens:
                param = AaveDailyParam(day, token)
                ret[param] = self.get_file_path(param)
        return ret

    def work(self):
        set_global_pbar(None)
        # if daily, global loop will handle processbar, outfile existence, gather param
        day_idx = self.from_config.start
        pbar = tqdm(total=(self.from_config.end - self.from_config.start).days + 1, ncols=80, position=0, leave=False)
        set_global_pbar(pbar)
        while day_idx <= self.from_config.end:
            # check file exist
            if self.config.to_config.skip_existed:
                all_exist = True
                for token in self.from_config.aave_config.tokens:
                    data_depends = AaveDailyParam(day_idx, token)
                    step_file_name = self.get_file_path(data_depends)
                    if not os.path.exists(step_file_name):
                        all_exist = False
                        break
                if all_exist:
                    day_idx += timedelta(days=1)
                    continue

            data_depends = {}
            for depend in self.depends:
                token_data = {}
                for token in self.from_config.aave_config.tokens:
                    path = depend.get_file_path(AaveDailyParam(day_idx, token))
                    token_data[token] = pd.read_csv(path, converters=depend.load_csv_converter)
                data_depends[depend.name] = token_data

            token_dfs = self._process_one_day(data_depends, day_idx, self.from_config.aave_config.tokens)
            for token, df in token_dfs.items():
                df.to_csv(self.get_file_path(AaveDailyParam(day_idx, token)), index=False)
            day_idx += timedelta(days=1)
            pbar.update()

    def _process_one_day(
        self, data: Dict[str, Dict[str, pd.DataFrame]], day: date, tokens: List[str]
    ) -> Dict[str, pd.DataFrame]:
        return {}



import json
import os
from datetime import datetime, timedelta, timezone
from decimal import Decimal

import pandas as pd
import requests

from demeter_fetch.common._typing import *


def get_file_name(chain: ChainType, pool_address, day: date | str):
    if isinstance(day, date):
        return f"{chain.name}-{pool_address}-{day.strftime('%Y-%m-%d')}.raw.csv"
    else:
        return f"{chain.name}-{pool_address}-{day}.raw.csv"


def get_aave_file_name(chain: ChainType, token_address, day: date):
    return f"{chain.name}-aave_v3-{token_address}-{day.strftime('%Y-%m-%d')}.raw.csv"


def convert_raw_file_name(file: str, to_config: ToConfig) -> str:
    file_name = os.path.basename(file)
    file_name_and_ext = os.path.splitext(file_name)

    return os.path.join(
        to_config.save_path,
        f"{file_name_and_ext[0].replace('.raw', '')}.{to_config.type.name}{file_name_and_ext[1]}",
    )


global_pbar = None


def set_global_pbar(pbar):
    global global_pbar
    global_pbar = pbar


def print_log(*args, tqdm_bar=None):
    if global_pbar is not None:
        tqdm_bar = global_pbar
    if tqdm_bar is not None:
        msg = str(*args).lstrip("(").rstrip(")")
        tqdm_bar.display(str(tqdm_bar) + f"    {str(msg)}")
        pass
    else:
        new_tuple = (datetime.now().strftime("%Y-%m-%d %H:%M:%S"),)
        new_tuple = new_tuple + args
        print(*new_tuple)


class TextUtil(object):
    @staticmethod
    def cut_after(text: str, symbol: str) -> str:
        index = text.find(symbol)
        return text[0:index]


class TimeUtil(object):
    @staticmethod
    def get_minute(time: datetime) -> datetime:
        return datetime(time.year, time.month, time.day, time.hour, time.minute, 0)

    @staticmethod
    def get_date_array(date_begin, date_end) -> List[date]:
        return [date_begin + timedelta(days=x) for x in range(0, 1 + (date_end - date_begin).days)]


class HexUtil(object):
    @staticmethod
    def to_signed_int(h):
        """
        Converts hex values to signed integers.
        """
        s = bytes.fromhex(h[2:])
        i = int.from_bytes(s, "big", signed=True)
        return i


def to_decimal(value):
    return Decimal(value) if value else Decimal(0)


class DataUtil(object):
    @staticmethod
    def fill_missing(data_list: List[MinuteData]) -> List[MinuteData]:
        if len(data_list) < 1:
            return data_list
        # take the first minute in data. instead of 0:00:00
        # so here will be a problem, if the first data is 0:03:00, the first 2 minutes will be blank
        # that's because there is no previous data to follow
        # those empty rows will be filled in loading stage
        index_minute = data_list[0].timestamp
        new_list = []
        data_list_index = 0

        start_day = data_list[0].timestamp.day
        while index_minute.day == start_day:
            if (data_list_index < len(data_list)) and (index_minute == data_list[data_list_index].timestamp):
                item = data_list[data_list_index]
                data_list_index += 1
            else:
                item = MinuteData()
                item.timestamp = index_minute
            prev_data = new_list[len(new_list) - 1] if len(new_list) - 1 >= 0 else None
            # if no previous(this might happen in the first minutes) data, this row will be discarded
            if item.fill_missing_field(prev_data):
                new_list.append(item)
            index_minute = index_minute + timedelta(minutes=1)

        return new_list


class ComplexEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime):
            return obj.strftime("%Y-%m-%d %H:%M:%S")
        elif isinstance(obj, date):
            return obj.strftime("%Y-%m-%d")
        else:
            return json.JSONEncoder.default(self, obj)


def hex_to_length(hex_str: str, new_length: int):
    """
    expend hex string to new length, will keep leading 0x if it exists
    eg:
    1. hex_str=0x1, length=3, result is 0x001
    2. hex_str=0x001, length=2, result is 0x01
    """

    has_0x = hex_str[0:2] == "0x" if len(hex_str) >= 2 else False

    hex_without_0x = hex_str if not has_0x else hex_str[2:]

    if len(hex_without_0x) <= new_length:
        return hex_without_0x.zfill(new_length) if not has_0x else "0x" + hex_without_0x.zfill(new_length)
    else:
        if hex_without_0x[-new_length - 1] != "0":
            raise RuntimeError("Not enough leading zeros to remove")
        return hex_without_0x[-new_length:] if not has_0x else "0x" + hex_without_0x[-new_length:]


def split_topic(value: str | list) -> List[str]:
    if isinstance(value, list):
        return value
    elif isinstance(value, str):
        value = value.strip("[]").replace('"', "").replace("'", "").replace(" ", "").replace("\n", ",")
        return value.split(",")
    else:
        raise RuntimeError("Unknown topic type")


def get_tx_type(topics_str):
    if not isinstance(topics_str, list) and pd.isna(topics_str):
        return topics_str
    topic_list = split_topic(topics_str)
    type_topic = topic_list[0]
    tx_type = KECCAK(type_topic)
    return tx_type


def get_transfer_from_logs(df: pd.DataFrame) -> pd.DataFrame:
    logs = df[["transaction_hash", "log_address", "topics", "data"]]
    logs["topics"] = logs["topics"].apply(split_topic)
    logs["topic0"] = logs["topics"].apply(lambda x: x[0])
    logs = logs[logs["topic0"] == KECCAK.TRANSFER.value]
    logs["from"] = logs["topics"].apply(lambda x: hex_to_length(x[1], 40))
    logs["to"] = logs["topics"].apply(lambda x: hex_to_length(x[2], 40))
    logs["value"] = logs["data"].apply(lambda x: Decimal(int(x, 16) if isinstance(x, str) else 0))
    logs = logs.drop(columns=["topics", "topic0", "data"])
    return logs


class ApiUtil:
    @staticmethod
    def query_blockno_from_time(
        chain: ChainType, blk_time: datetime, is_before: bool = True, proxy="", etherscan_api_key=None
    ):
        proxies = (
            {
                "http": proxy,
                "https": proxy,
            }
            if proxy
            else {}
        )
        before_or_after = "before" if is_before else "after"
        url = ChainTypeConfig[chain]["query_height_api"]
        blk_time = blk_time.replace(tzinfo=timezone.utc)
        url = url.replace("%1", str(int(blk_time.timestamp()))).replace("%2", before_or_after)
        if etherscan_api_key is not None:
            url += "&apikey=" + etherscan_api_key
        retry = 3
        result = None
        while retry:
            try:
                result = requests.get(url, proxies=proxies)
                retry = 0
            except Exception as ex:
                print("Query etherscan failed, error: " + str(ex))
                retry -= 1
        if not result:
            raise RuntimeError("request error with retry 3 times.")
        if result.status_code != 200:
            raise RuntimeError("request block number failed, code: " + str(result.status_code))
        result_json = result.json()
        if int(result_json["status"]) == 1:
            return int(result_json["result"])
        else:
            raise RuntimeError("request block number failed, message: " + str(result_json))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 2024-01-08 16:37
# @Author  : 32ethers
# @Description:

from demeter_fetch.common import *


def get_item_with_default_2(cfg: Dict, key1: str, key2: str, default_val, converter=None):
    val = default_val
    if key2 in cfg[key1]:
        val = cfg[key1][key2]
    if converter is None or val is None:
        return val
    else:
        return converter(val)


def get_item_with_default_3(cfg: Dict, key1: str, key2: str, key3: str, default_val, converter=None):
    val = default_val
    if key3 in cfg[key1][key2]:
        val = cfg[key1][key2][key3]
    if converter is None or val is None:
        return val
    else:
        return converter(val)


def convert_to_config(conf_file: dict) -> Config:
    to_type = ToType[conf_file["to"]["type"]]
    save_path = get_item_with_default_2(conf_file, "to", "save_path", "../")
    multi_process = get_item_with_default_2(conf_file, "to", "multi_process", False)
    skip_existed = get_item_with_default_2(conf_file, "to", "skip_existed", False)
    keep_raw = get_item_with_default_2(conf_file, "to", "keep_raw", False)
    to_config = ToConfig(to_type, save_path, multi_process, skip_existed, keep_raw)

    chain = ChainType[conf_file["from"]["chain"]]
    data_source = DataSource[conf_file["from"]["datasource"]]
    dapp_type = DappType[conf_file["from"]["dapp_type"]]
    http_proxy = get_item_with_default_2(conf_file, "from", "http_proxy", None)
    start_time = get_item_with_default_2(
        conf_file, "from", "start", None, lambda x: datetime.strptime(x, "%Y-%m-%d").date()
    )
    end_time = get_item_with_default_2(
        conf_file, "from", "end", None, lambda x: datetime.strptime(x, "%Y-%m-%d").date()
    )
    from_config = FromConfig(
        chain=chain, data_source=data_source, dapp_type=dapp_type, http_proxy=http_proxy, start=start_time, end=end_time
    )
    if start_time is None or end_time is None:
        raise RuntimeError("start time and end time must be set")
    if dapp_type == DappType.uniswap:
        pool_address = conf_file["from"]["uniswap"]["pool_address"].lower()
        ignore_position_id = get_item_with_default_3(conf_file, "from", "uniswap", "ignore_position_id", False)
        from_config.uniswap_config = UniswapConfig(pool_address, ignore_position_id)

    elif dapp_type == DappType.aave:
        token_addresses = [x.lower() for x in conf_file["from"]["aave"]["tokens"]]
        from_config.aave_config = AaveConfig(token_addresses)

    if data_source not in ChainTypeConfig[from_config.chain]["allow"]:
        raise RuntimeError(f"{data_source.name} is not allowed to download from {from_config.chain.name}")
    match data_source:
        case DataSource.rpc:
            if "rpc" not in conf_file["from"]:
                raise RuntimeError("should have [from.rpc]")
            auth_string = get_item_with_default_3(conf_file, "from", "rpc", "auth_string", None)
            keep_tmp_files = get_item_with_default_3(conf_file, "from", "rpc", "keep_tmp_files", None)
            if keep_raw == False and keep_tmp_files is None:
                keep_tmp_files = False
            etherscan_api_key = get_item_with_default_3(conf_file, "from", "rpc", "etherscan_api_key", None)
            end_point = conf_file["from"]["rpc"]["end_point"]
            batch_size = get_item_with_default_3(conf_file, "from", "rpc", "batch_size", 500)
            force_no_proxy = get_item_with_default_3(conf_file, "from", "rpc", "force_no_proxy", False)
            from_config.rpc = RpcConfig(
                end_point=end_point,
                batch_size=batch_size,
                auth_string=auth_string,
                keep_tmp_files=keep_tmp_files,
                etherscan_api_key=etherscan_api_key,
                force_no_proxy=force_no_proxy,
            )
        case DataSource.big_query:
            if "big_query" not in conf_file["from"]:
                raise RuntimeError("should have [from.big_query]")
            auth_file = conf_file["from"]["big_query"]["auth_file"]
            from_config.big_query = BigQueryConfig(
                auth_file=auth_file,
            )
        case DataSource.chifra:
            if "chifra" not in conf_file["from"]:
                raise RuntimeError("should have [from.chifra]")
            etherscan_api_key = get_item_with_default_3(conf_file, "from", "chifra", "etherscan_api_key", None)
            from_config.chifra_config = ChifraConfig(
                etherscan_api_key=etherscan_api_key,
            )

    return Config(from_config, to_config)



import datetime
from dataclasses import dataclass
from decimal import Decimal
from typing import Dict, Callable

import pandas as pd

from .uniswap_utils import match_proxy_log, handle_event, handle_proxy_event
from ..common import to_decimal, DailyNode, UniNodesNames, DailyParam, get_tx_type


@dataclass
class PoolTick:
    block_number = 0
    block_timestamp = 0
    tx_type = 0
    transaction_hash = 0
    pool_tx_index = 0
    pool_log_index = 0
    proxy_log_index = 0
    sender = 0
    receipt = 0
    amount0 = 0
    amount1 = 0
    total_liquidity = 0
    total_liquidity_delta = 0
    sqrtPriceX96 = 0
    current_tick = 0
    position_id = 0
    tick_lower = 0
    tick_upper = 0
    liquidity = 0


def convert_pool_tick_df(input_df: pd.DataFrame) -> pd.DataFrame:
    df = input_df.copy()
    df["tx_type"] = df.apply(lambda x: get_tx_type(x.topics), axis=1)
    df[
        [
            "sender",
            "receipt",
            "amount0",
            "amount1",
            "sqrtPriceX96",
            "total_liquidity",
            "current_tick",
            "tick_lower",
            "tick_upper",
            "liquidity",
            "total_liquidity_delta",
        ]
    ] = df.apply(lambda r: handle_event(r.tx_type, r.topics, r.data), axis=1, result_type="expand")
    df = df.drop(columns=["topics", "data"])
    df = df.sort_values(["block_number", "log_index"], ascending=[True, True])
    df[["sqrtPriceX96", "total_liquidity", "current_tick"]] = df[
        ["sqrtPriceX96", "total_liquidity", "current_tick"]
    ].ffill()

    df["total_liquidity_delta"] = df["total_liquidity_delta"].fillna(0)
    # convert type to keep decimal
    df["sqrtPriceX96"] = df.apply(lambda x: convert_to_decimal(x.sqrtPriceX96), axis=1)
    df["total_liquidity_delta"] = df.apply(lambda x: convert_to_decimal(x.total_liquidity_delta), axis=1)
    df["liquidity"] = df.apply(lambda x: convert_to_decimal(x.liquidity), axis=1)
    df["total_liquidity"] = df.apply(lambda x: convert_to_decimal(x.total_liquidity), axis=1)

    df["total_liquidity_delta"] = df.apply(
        lambda x: handle_tick(x.tick_lower, x.tick_upper, x.current_tick, x.total_liquidity_delta),
        axis=1,
    )
    df["total_liquidity"] = df.apply(lambda x: x.total_liquidity_delta + x.total_liquidity, axis=1)
    df["tx_type"] = df.apply(lambda x: x.tx_type.name, axis=1)
    df = df.rename(columns={"transaction_index": "tx_index"})
    order = [
        "block_number",
        "block_timestamp",
        "tx_type",
        "transaction_hash",
        "tx_index",
        "log_index",
        "sender",
        "receipt",
        "amount0",
        "amount1",
        "total_liquidity",
        "total_liquidity_delta",
        "sqrtPriceX96",
        "current_tick",
        "tick_lower",
        "tick_upper",
        "liquidity",
    ]
    df = df[order]

    return df


class UniTick(DailyNode):
    def __init__(self, depends):
        super().__init__(depends)
        self.name = UniNodesNames.tick

    def _get_file_name(self, param: DailyParam) -> str:
        return f"{self.from_config.chain.name}-{self.from_config.uniswap_config.pool_address}-{param.day.strftime('%Y-%m-%d')}.tick.csv"

    @property
    def load_csv_converter(self) -> Dict[str, Callable]:
        return {
            "amount0": to_decimal,
            "amount1": to_decimal,
            "liquidity": to_decimal,
            "total_liquidity": to_decimal,
            "total_liquidity_delta": to_decimal,
            "sqrtPriceX96": to_decimal,
        }

    def _process_one_day(self, data: Dict[str, pd.DataFrame], day: datetime.date) -> pd.DataFrame:
        pool_df = data[UniNodesNames.pool]
        proxy_df = data[UniNodesNames.proxy_lp]
        match_proxy_log(pool_df, proxy_df)
        pool_df = pool_df.sort_values(["block_number", "log_index"], ascending=[True, True])

        merged_df = convert_pool_tick_df(pool_df)
        merged_df[["proxy_topics", "proxy_data", "proxy_log_index"]] = pool_df[
            ["proxy_topics", "proxy_data", "proxy_log_index"]
        ]
        merged_df.rename(columns={"tx_index": "pool_tx_index", "log_index": "pool_log_index"}, inplace=True)

        merged_df["position_id"] = merged_df.apply(lambda x: handle_proxy_event(x.proxy_topics), axis=1)
        order = [
            "block_number",
            "block_timestamp",
            "tx_type",
            "transaction_hash",
            "pool_tx_index",
            "pool_log_index",
            "proxy_log_index",
            "sender",
            "receipt",
            "amount0",
            "amount1",
            "total_liquidity",
            "total_liquidity_delta",
            "sqrtPriceX96",
            "current_tick",
            "position_id",
            "tick_lower",
            "tick_upper",
            "liquidity",
        ]
        merged_df = merged_df[order]
        return merged_df


class UniTickNoPos(DailyNode):
    def __init__(self, depends):
        super().__init__(depends)
        self.name = UniNodesNames.tick_without_pos

    def _get_file_name(self, param: DailyParam) -> str:
        return f"{self.from_config.chain.name}-{self.from_config.uniswap_config.pool_address}-{param.day.strftime('%Y-%m-%d')}.pool.tick.csv"

    @property
    def load_csv_converter(self) -> Dict[str, Callable]:
        return {
            "amount0": to_decimal,
            "amount1": to_decimal,
            "total_liquidity": to_decimal,
            "total_liquidity_delta": to_decimal,
            "sqrtPriceX96": to_decimal,
        }

    def _process_one_day(self, data: Dict[str, pd.DataFrame], day: datetime.date) -> pd.DataFrame:
        input_param = data[UniNodesNames.pool]
        df = convert_pool_tick_df(input_param)
        return df


def convert_to_decimal(value):
    return Decimal(value) if value else Decimal(0)


def handle_tick(lower_tick, upper_tick, current_tick, delta):
    if lower_tick is None or upper_tick is None or current_tick is None:
        return 0
    if lower_tick < current_tick < upper_tick:
        return delta
    else:
        return 0



import argparse


def get_commend_args():
    argParser = argparse.ArgumentParser()
    argParser.add_argument("-c", "--config", help="Path of config file")
    parser_tool_sub = argParser.add_subparsers(help="demeter-fetch tools", dest="tools")

    parser_tool_chifra = parser_tool_sub.add_parser("date_to_height", help="Query block height from etherscan in certain date range")
    parser_tool_chifra.add_argument("-c", "--chain", help="chain name, [ethereum, polygon]")
    parser_tool_chifra.add_argument("-s", "--start", help="start date, eg: 2023-1-1")
    parser_tool_chifra.add_argument("-e", "--end", help="end date, eg: 2023-1-1")
    parser_tool_chifra.add_argument("-p", "--http_proxy", help="proxy, eg: https://localhost:7890, optional", default="")
    parser_tool_chifra.add_argument("-k", "--key", help="key in etherscan, optional", default="")

    args = argParser.parse_args()
    return args



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 2024-01-08 14:22
# @Author  : 32ethers
# @Description:
from typing import List

from ..common import Node
from ..processor_aave import AaveMinute, AaveTick
from ..processor_uniswap import UniUserLP, UniPositions, UniTick, UniTickNoPos, UniMinute

from ..sources import UniSourcePool, UniSourceProxyTransfer, UniSourceProxyLp, AaveSource, UniTransaction
from .. import DappType, ToType, UniNodesNames, AaveNodesNames


def _get_reversed_copy(list_to_reverse):
    ret_list = list_to_reverse.copy()
    ret_list.reverse()
    return ret_list


def get_relative_nodes(root: Node) -> List[Node]:
    depth_first_array = []
    stack = [root]
    while len(stack) > 0:
        poped: Node = stack.pop()
        depth_first_array.append(poped)
        if len(poped.depends) > 0:
            # if you want to use a list as a stack, you have to add elements in the tail
            stack.extend(_get_reversed_copy(poped.depends))
    depth_first_array.reverse()
    no_dup_array = []
    for n in depth_first_array:
        if n not in no_dup_array:
            no_dup_array.append(n)
    return no_dup_array


class UniNodes:
    pool = UniSourcePool([])
    proxy_transfer = UniSourceProxyTransfer([])
    proxy_lp = UniSourceProxyLp([])
    minute = UniMinute([pool])
    tick = UniTick([pool, proxy_lp])
    tick_without_pos = UniTickNoPos([pool])
    tx = UniTransaction([tick])
    position = UniPositions([tick, tx])
    user_lp = UniUserLP([position])


class AaveNodes:
    raw = AaveSource([])

    minute = AaveMinute([raw])
    tick = AaveTick([raw])


def get_root_node(dapp: DappType, to_type: ToType, ignore_pos_id: bool = False) -> Node:
    if dapp == DappType.uniswap:
        match to_type:
            case ToType.raw:
                return UniNodes.pool
            case ToType.tick:
                if ignore_pos_id:
                    return UniNodes.tick_without_pos
                else:
                    return UniNodes.tick
            case ToType.position:
                return UniNodes.position
            case ToType.minute:
                return UniNodes.minute
            case ToType.user_lp:
                return UniNodes.user_lp
            case _:
                raise NotImplemented(f"{dapp} {to_type} not supported")

    elif dapp == DappType.aave:
        match to_type:
            case ToType.raw:
                return AaveNodes.raw
            case ToType.minute:
                return AaveNodes.minute
            case ToType.tick:
                return AaveNodes.tick
            case _:
                raise NotImplemented(f"{dapp} {to_type} not supported")
    else:
        raise NotImplemented(f"{dapp} not supported")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 2024-01-08 11:31
# @Author  : 32ethers
# @Description:

from .commands import get_commend_args
from .downloader import download
from .engine import get_relative_nodes



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 2024-01-08 16:21
# @Author  : 32ethers
# @Description:

import dataclasses
import json
import os
from typing import List

import toml

import demeter_fetch.common as utils
from . import engine
from .config import convert_to_config
from ..common import print_log, set_global_pbar, Node


def download(cfg_path):
    if not os.path.exists(cfg_path):
        utils.print_log("config file not found,")
        exit(1)
    config_file = toml.load(cfg_path)
    try:
        config = convert_to_config(config_file)
    except RuntimeError as e:
        utils.print_log(e)
        exit(1)
    utils.print_log(
        "Download config:",
        json.dumps(dataclasses.asdict(config), cls=utils.ComplexEncoder, indent=4),
    )

    ignore_pos = False
    if config.from_config.uniswap_config is not None:
        ignore_pos = config.from_config.uniswap_config.ignore_position_id

    root_step = engine.get_root_node(config.from_config.dapp_type, config.to_config.type, ignore_pos)
    steps: List[Node] = engine.get_relative_nodes(root_step)
    utils.print_log("Will execute the following steps: ", steps)
    [step.set_config(config) for step in steps]

    for step in steps:
        set_global_pbar(None)
        print_log(f"Current step: {step.name}")
        step.work()
    # remove raw files
    if not config.to_config.keep_raw:
        for step in steps:
            if step != root_step:
                for param, sf in step.get_file_paths.items():
                    os.remove(sf)
